/* gwavefile.c: read wave data directly from a .WAV file
 * Copyright (c) 2000 David A. Bartold
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <stdio.h>
#include "gtkwavebuffer.h"
#include "gwavefile.h"

#if 0
typedef struct _GtkFastwave GtkFastwave;

struct _GtkFastwave
{
  FILE           *file;

  guint32         rate;
  guint32         n_channels;
  guint32         length;
  guint32         width;
  GtkWaveBufferType  datatype;
  guint32         data_offset;
};
#endif


GWavefile      *g_wavefile_open             (gchar     *file);
void            g_wavefile_close            (GWavefile *wave);
GtkWaveBufferType  g_wavefile_get_datatype  (GWavefile *wave);
guint32         g_wavefile_get_length       (GWavefile *wave);
guint32         g_wavefile_get_num_channels (GWavefile *wave);
guint32         g_wavefile_get_rate         (GWavefile *wave);
void            g_wavefile_read             (GWavefile *wave,
                                             guint32    offset,
                                             guint32    length,
                                             gpointer   buf);

#define WAVEID(a,b,c,d) (((a) << 24) | ((b) << 16) | ((c) << 8) | d)

static guint32
get_id (FILE *in)
{
  guint32 v;

  v =  fgetc (in) << 24;
  v |= fgetc (in) << 16;
  v |= fgetc (in) << 8;
  v |= fgetc (in);

  return v;
}

static guint32
get_32 (FILE *in)
{
  guint32 v;

  v =  fgetc (in);
  v |= fgetc (in) << 8;
  v |= fgetc (in) << 16;
  v |= fgetc (in) << 24;

  return v;
}

static guint16
get_16 (FILE *in)
{
  guint32 v;

  v =  fgetc (in);
  v |= fgetc (in) << 8;

  return v;
}

GWavefile *
g_wavefile_open (gchar *file)
{
  GWavefile *wave;
  FILE *f;
  guint32 status;
  guint32 file_size, chunk_size, pos, id, rate;
  guint32 found_fmt, found_data;

  guint32 data_length, data_offset;
  guint32 fmt_tag, fmt_channels, fmt_blockalign, fmt_bits;
  GtkWaveBufferType fmt_datatype;

  status = 0;
  f = fopen (file, "r");
  if (f == NULL)
    return NULL;

  if (get_id (f) != WAVEID ('R', 'I', 'F', 'F'))
    status = 1;

  if (status == 0)
    {
      file_size = get_32 (f);

      if (status == 0)
        if (get_id (f) != WAVEID ('W', 'A', 'V', 'E'))
          status = 1;
    }

  if (status == 0)
    {
      file_size += 8;

      pos = 12;
      found_fmt = 0;
      found_data = 0;
      while (pos < file_size)
        {
          id = get_id (f);
          chunk_size = get_32 (f);

          switch (id)
            {
              case WAVEID ('f', 'm', 't', ' '):
                if (found_fmt == 1)
                   status = 1;
                else
                  {
                    fmt_tag = get_16 (f);
                    fmt_channels = get_16 (f);
                    rate = get_32 (f);
                    get_32 (f);
                    fmt_blockalign = get_16 (f);
                    fmt_bits = get_16 (f);
                    found_fmt = 1;
                  }
                break;

              case WAVEID ('d', 'a', 't', 'a'):
                if (found_data == 1)
                   status = 1;
                else
                  {
                    data_length = chunk_size;
                    data_offset = pos + 8;  
                    found_data = 1;
                  }
                break;
            }

          pos += chunk_size + 8;
          fseek (f, pos, SEEK_SET);
        }

      if (found_fmt == 0 || found_data == 0)
        status = 1;
    }

  if (status == 0)
    {
      if (fmt_bits == 8)
        fmt_datatype = GTK_WAVE_BUFFER_TYPE_U8;
      else if (fmt_bits == 16)
        fmt_datatype = GTK_WAVE_BUFFER_TYPE_S16;
      else if (fmt_bits == 32)
        fmt_datatype = GTK_WAVE_BUFFER_TYPE_S32;
      else
        status = 1;
    }

  if (status == 0)
    status = ferror (f);

  if (status == 0)
    {
      wave = g_new (GWavefile, 1);
      wave->file = f;
      wave->datatype = fmt_datatype;
      wave->rate = rate;
      wave->n_channels = fmt_channels;
      wave->width = gtk_wave_buffer_type_width (wave->datatype) * wave->n_channels;
      wave->length = data_length / wave->width;
      wave->data_offset = data_offset;

      return wave;
    }

  if (f != NULL)
    fclose (f);

  return NULL;
}

void
g_wavefile_close (GWavefile *wave)
{
  fclose (wave->file);
  g_free (wave);
}

GtkWaveBufferType
g_wavefile_get_datatype (GWavefile *wave)
{
  return wave->datatype;
}

guint32
g_wavefile_get_length (GWavefile *wave)
{
  return wave->length;
}

guint32
g_wavefile_get_num_channels (GWavefile *wave)
{
  return wave->n_channels;
}

guint32
g_wavefile_get_rate (GWavefile *wave)
{
  return wave->rate;
}

void
g_wavefile_read (GWavefile *wave, guint32 offset, guint32 length, gpointer buf)
{
  g_return_if_fail (offset + length <= wave->length);

  fseek (wave->file, wave->data_offset + offset * wave->width, SEEK_SET);
  fread (buf, wave->width, length, wave->file);
}
