//////////////////////////////////////////////////////////////////////
//  guileconfig.cc
//
//  Copyright 1999,2000 Bill Gribble <grib@billgribble.com>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  $Id: guileconfig.cc,v 1.6 2000/10/04 16:45:31 pbd Exp $
//
//////////////////////////////////////////////////////////////////////

#include <string>

#include <unistd.h>
#include <guile/gh.h>
#include <vector>
#include <pbd/error.h>

#include <guileconfig/guileconfig.h>

guileconfig * guileconfig::Config = NULL;

//////////////////////////////////////////////////////////////////////
//  configuration variable setters 
//
//  config vars are set from scheme, through set-config-var.
//  the C types that can be set are int, float, bool, char * (string), 
//  char ** (vector of string), guileconfig::AudioDeviceRequest (soundcard on the 
//  scheme side), and MIDI::PortRequest (mididev on the scheme side).
//////////////////////////////////////////////////////////////////////

SCM
set_integer_c_config_variable(SCM var_name, SCM value) {
  char  * tag;
  int   val;
  int   len;

  tag  = gh_scm2newstr(var_name, &len);
  val = gh_scm2long(value);

  if(guileconfig::Config->variables[tag]) {
    *(int *)(guileconfig::Config->variables[tag]) = val;      
    return gh_bool2scm(1);
  }
  else {
    return gh_bool2scm(0);
  }
}

SCM
set_real_c_config_variable(SCM var_name, SCM value) {
  char   * tag;
  float  val;
  int    len;

  tag  = gh_scm2newstr(var_name, &len);
  val = (float)gh_scm2double(value);
  
  if(guileconfig::Config->variables[tag]) {
    *(float *)(guileconfig::Config->variables[tag]) = val;  
    return gh_bool2scm(1);
  }
  else {
    return gh_bool2scm(0);
  }
}

SCM
set_boolean_c_config_variable(SCM var_name, SCM value) {
  char   * tag;
  bool   val;
  int    len; 

  tag  = gh_scm2newstr(var_name, &len);
  val = (bool)gh_scm2bool(value);
  
  if(guileconfig::Config->variables[tag]) {
    *(bool *)(guileconfig::Config->variables[tag]) = val;  
    return gh_bool2scm(1);
  }
  else {
    return gh_bool2scm(0);
  }
}


SCM
set_string_c_config_variable(SCM var_name, SCM value) {
  char   * tag;
  char   * val;
  int    len;

  tag  = gh_scm2newstr(var_name, &len);
  val =  gh_scm2newstr(value, &len);

  if(guileconfig::Config->variables[tag]) {
    *((string *)(guileconfig::Config->variables[tag])) = string(val);  
    return gh_bool2scm(1);
  }
  else {
    return gh_bool2scm(0);
  }
}

SCM
set_charptr_c_config_variable(SCM var_name, SCM value) {
  char   * tag;
  char   * val;
  int    len;

  tag  = gh_scm2newstr(var_name, &len);
  val =  gh_scm2newstr(value, &len);

  if(guileconfig::Config->variables[tag]) {
    *(char **)(guileconfig::Config->variables[tag]) = val;  
    return gh_bool2scm(1);
  }
  else {
    return gh_bool2scm(0);
  }
}

SCM
set_charptrlist_c_config_variable(SCM var_name, SCM value) {
  char   * tag;
  int    list_length;
  SCM    length_proc;
  char   ** stringlist;
  int    len;

  // value is a list of strings.  We want to make a C++ vector 
  // of pointers to the strings. 
  tag  = gh_scm2newstr(var_name, &len);
  length_proc = gh_eval_str("length");
  list_length = gh_scm2int(gh_call1(length_proc, value));
  
  stringlist = new (char *)[list_length+1];
  for(int i = 0; i < list_length; i++) {
    stringlist[i] = gh_scm2newstr(gh_car(value), &len);
    value = gh_cdr(value);
  }
  stringlist[list_length] = 0;

  if(guileconfig::Config->variables[tag]) {
    *(char ***)(guileconfig::Config->variables[tag]) = stringlist;  
    return gh_bool2scm(1);
  }
  else {
    return gh_bool2scm(0);
  }
}


SCM
set_mididevlist_c_config_variable(SCM var_name, SCM value) {
  char                        * tag;
  int                         list_length;
  SCM                         length_proc;
  SCM                         slots;
  SCM                         retval = value; 
  guileconfig::MidiDeviceRequest           * req;
  vector<guileconfig::MidiDeviceRequest *> * reqlist;
  int                         len;

  // value is a list of MIDI Port requests.  We want to stuff them
  // into a C++ vector of pointers to port requests.  
  tag  = gh_scm2newstr(var_name, &len);
  length_proc = gh_eval_str("length");
  list_length = gh_scm2int(gh_call1(length_proc, value));
  
  reqlist = (vector<guileconfig::MidiDeviceRequest *> *)
    guileconfig::Config->variables[tag];

  if(reqlist == NULL) {
    return gh_bool2scm(0);
  }
  
  reqlist->erase(reqlist->begin(), reqlist->end());
  
  for (int i = 0; i < list_length; i++) {
    req   = new guileconfig::MidiDeviceRequest;
    
    // the simple-object slots are in the cdr of each object,
    // and the current object is the head of the list, so 
    // we are saying (set! slots (cdr (car value)))
    slots = gh_cdar(value); 
    
    req->device = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(0)), &len);
    req->tag = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(1)), &len);
    req->type    = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(2)), &len);
    req->mode    = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(3)), &len);
    
    reqlist->push_back(req);
    value = gh_cdr(value);
  }
  return retval;
}


SCM
set_soundcard_c_config_variable(SCM var_name, SCM value) {
  char * tag;
  guileconfig::AudioDeviceRequest * val;
  SCM  slots;
  int    len;
  
  // get the tag and value
  tag   = gh_scm2newstr(var_name, &len);
  val   = (guileconfig::AudioDeviceRequest *)(guileconfig::Config->variables[tag]);
  
  // pick the slots vector out of the simple-object 
  slots = gh_cdr(value);

  if(guileconfig::Config->variables[tag]) {
    val->device = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(0)), &len);
    val->mode    = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(1)), &len);
    val->type    = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(2)), &len);
    val->format  = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(3)), &len);
    val->rate    =  gh_scm2int(gh_vector_ref(slots, gh_int2scm(4)));
    val->nchannels     =  gh_scm2int(gh_vector_ref(slots, gh_int2scm(5)));
    val->latency = (float)gh_scm2double(gh_vector_ref(slots, gh_int2scm(6)));

    return gh_bool2scm(1);
  }
  else {
    return gh_bool2scm(0);
  }
}

SCM
set_mididev_c_config_variable(SCM var_name, SCM value) {
  char * tag;
  guileconfig::MidiDeviceRequest * val;
  SCM  slots;
  int    len;
  
  // get the tag and value
  tag   = gh_scm2newstr(var_name, &len);
  val   = (guileconfig::MidiDeviceRequest *)(guileconfig::Config->variables[tag]);

  // pick the slots vector out of the simple-object 
  slots = gh_cdr(value);

  if(guileconfig::Config->variables[tag]) {
    val->device  = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(0)), &len);
    val->tag     = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(1)), &len);
    val->type    = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(2)), &len);
    val->mode    = gh_scm2newstr(gh_vector_ref(slots, gh_int2scm(3)), &len);
    return gh_bool2scm(1);
  }
  else {
    return gh_bool2scm(0);
  }
}

//////////////////////////////////////////////////////////////////////
//  config variable getters 
//////////////////////////////////////////////////////////////////////

SCM
get_integer_c_config_variable(SCM var_name) {
  char  * tag;
  int   len;
  void  * vptr;
  SCM   retval;

  tag  = gh_scm2newstr(var_name, &len);
  vptr = guileconfig::Config->variables[tag];
  
  if(vptr) {
    retval = gh_int2scm(*(int *)vptr);
  }
  else {
    retval = gh_bool2scm(0);
  }
  return retval;
}

SCM
get_real_c_config_variable(SCM var_name) {
  char  * tag;
  int   len;
  void  * vptr;
  SCM   retval;

  tag  = gh_scm2newstr(var_name, &len);
  vptr = guileconfig::Config->variables[tag];
  
  if(vptr) {
    retval = gh_double2scm(*(float *)vptr);
  }
  else {
    retval = gh_bool2scm(0);
  }
  return retval;
}


SCM
get_boolean_c_config_variable(SCM var_name) {
  char  * tag;
  int   len;
  void  * vptr;
  SCM   retval;

  tag  = gh_scm2newstr(var_name, &len);
  vptr = guileconfig::Config->variables[tag];
  
  if(vptr) {
    retval = gh_bool2scm(*(bool *)vptr);
  }
  else {
    retval = gh_bool2scm(0);
  }
  return retval;
}

SCM
get_string_c_config_variable(SCM var_name) {
  char  * tag;
  int   len;
  void  * vptr;
  SCM   retval;

  tag  = gh_scm2newstr(var_name, &len);
  vptr = guileconfig::Config->variables[tag];
  
  if(vptr) {
    retval = gh_str02scm((char *)((string *)vptr)->c_str());
  }
  else {
    retval = gh_bool2scm(0);
  }
  return retval;
}

SCM
get_charptr_c_config_variable(SCM var_name) {
  char  * tag;
  int   len;
  void  * vptr;
  SCM   retval;

  tag  = gh_scm2newstr(var_name, &len);
  vptr = guileconfig::Config->variables[tag];
  
  if(vptr) {
    retval = gh_str02scm((char *)vptr);
  }
  else {
    retval = gh_bool2scm(0);
  }
  return retval;
}


SCM
get_charptrlist_c_config_variable(SCM var_name) {
  char  * tag;
  int   len;
  void  * vptr;
  char  ** cur_string;
  SCM   retval = SCM_EOL;
  SCM   one_string;

  tag  = gh_scm2newstr(var_name, &len);
  vptr = guileconfig::Config->variables[tag];
  cur_string = *(char ***)vptr;

  if(vptr) {
    while(*cur_string) {
      one_string = gh_str02scm(*cur_string);
      retval = gh_cons(one_string, retval);
      cur_string++;
    }
    retval = gh_reverse(retval);
  }
  else {
    retval = gh_bool2scm(0);
  }
  return retval;
}

SCM
get_mididevlist_c_config_variable(SCM var_name) {
  char  * tag;
  int   len;
  SCM   retval  = SCM_EOL;
  SCM   typetag = gh_symbol2scm("mididev");
  SCM   slots;
  vector<guileconfig::MidiDeviceRequest *> * reqptr;
  vector<guileconfig::MidiDeviceRequest *>::iterator it;

  tag  = gh_scm2newstr(var_name, &len);
  reqptr = (vector<guileconfig::MidiDeviceRequest *> *)
    (guileconfig::Config->variables[tag]);
  
  if(reqptr) {
    for(it = reqptr->begin(); it != reqptr->end(); it++) {
      slots = gh_make_vector(gh_int2scm(4), SCM_BOOL_F);

      gh_vector_set_x(slots, gh_int2scm(0), 
                      gh_str02scm((char *) (*it)->device.c_str()));
      gh_vector_set_x(slots, gh_int2scm(1), 
                      gh_str02scm((char *) (*it)->tag.c_str()));
      gh_vector_set_x(slots, gh_int2scm(2), 
		      gh_str02scm((char *) (*it)->type.c_str()));
      gh_vector_set_x(slots, gh_int2scm(3), 
		      gh_str02scm((char *) (*it)->mode.c_str()));
      retval = gh_cons(gh_cons(typetag, slots), retval);
    }
    return retval;
  }
  else {
    return SCM_BOOL_F;
  }
}

SCM
get_soundcard_c_config_variable(SCM var_name) {
  char  * tag;
  int   len;
  guileconfig::AudioDeviceRequest * reqptr;
  SCM   retval;
  SCM   typetag = gh_symbol2scm("soundcard");
  SCM   slots = gh_make_vector(gh_int2scm(7), gh_bool2scm(0));

  tag  = gh_scm2newstr(var_name, &len);
  reqptr = (guileconfig::AudioDeviceRequest *)(guileconfig::Config->variables[tag]);
  
  if(reqptr) {
    gh_vector_set_x(slots, gh_int2scm(0), 
                    gh_str02scm((char *) reqptr->device.c_str()));
    gh_vector_set_x(slots, gh_int2scm(1), 
		    gh_str02scm((char *) reqptr->mode.c_str()));
    gh_vector_set_x(slots, gh_int2scm(2), 
		    gh_str02scm((char *) reqptr->type.c_str()));
    gh_vector_set_x(slots, gh_int2scm(3),
		    gh_str02scm((char *) reqptr->format.c_str()));
    gh_vector_set_x(slots, gh_int2scm(4), gh_int2scm(reqptr->rate));
    gh_vector_set_x(slots, gh_int2scm(5), gh_int2scm(reqptr->nchannels));
    gh_vector_set_x(slots, gh_int2scm(6), gh_double2scm((double)reqptr->latency)); 
    retval = gh_cons(typetag, slots);
  }
  else {
    retval = gh_bool2scm(0);
  }
  return retval;
}

SCM
get_mididev_c_config_variable(SCM var_name) {
  char  * tag;
  int   len;
  guileconfig::MidiDeviceRequest * reqptr;
  SCM   retval;
  SCM   typetag = gh_symbol2scm("mididev");
  SCM   slots = gh_make_vector(gh_int2scm(4), gh_bool2scm(0));

  tag  = gh_scm2newstr(var_name, &len);
  reqptr = (guileconfig::MidiDeviceRequest *)(guileconfig::Config->variables[tag]);
  
  if(reqptr) {
    gh_vector_set_x(slots, gh_int2scm(0), 
		    gh_str02scm((char *) reqptr->device.c_str()));
    gh_vector_set_x(slots, gh_int2scm(1), 
		    gh_str02scm((char *) reqptr->tag.c_str()));
    gh_vector_set_x(slots, gh_int2scm(2), 
		    gh_str02scm((char *) reqptr->type.c_str()));
    gh_vector_set_x(slots, gh_int2scm(3), 
		    gh_str02scm((char *) reqptr->mode.c_str()));
    retval = gh_cons(typetag, slots);
  }
  else {
    retval = gh_bool2scm(0);
  }
  return retval;
}


//////////////////////////////////////////////////////////////////////
//  guileconfig::guileconfig
//  
//  creating a guileconfig object causes all these functions to
//  get published to scheme.  We have to have a separate bootstrap
//  function to let the derived class constructor finish before 
//  actually loading user code. 
//////////////////////////////////////////////////////////////////////

guileconfig::guileconfig(const char *appname) 
	: _app_name (strdup (appname))
{
		
	gh_new_procedure2_0("%set-integer-c-config-variable",
			    set_integer_c_config_variable);
	
	gh_new_procedure2_0("%set-real-c-config-variable",
			    set_real_c_config_variable);
	
	gh_new_procedure2_0("%set-boolean-c-config-variable",
			    set_boolean_c_config_variable);
	
	gh_new_procedure2_0("%set-soundcard-c-config-variable",
			    set_soundcard_c_config_variable);
	
	gh_new_procedure2_0("%set-mididev-c-config-variable",
			    set_mididev_c_config_variable);
	
	gh_new_procedure2_0("%set-mididevlist-c-config-variable",
			    set_mididevlist_c_config_variable);
	
	gh_new_procedure2_0("%set-string-c-config-variable",
			    set_string_c_config_variable);
	
	gh_new_procedure2_0("%set-charptr-c-config-variable",
			    set_charptr_c_config_variable);
	
	gh_new_procedure2_0("%set-charptrlist-c-config-variable",
			    set_charptrlist_c_config_variable);
	
	gh_new_procedure1_0("%get-integer-c-config-variable",
			    get_integer_c_config_variable);
	
	gh_new_procedure1_0("%get-real-c-config-variable",
			    get_real_c_config_variable);
	
	gh_new_procedure1_0("%get-boolean-c-config-variable",
			    get_boolean_c_config_variable);
	
	gh_new_procedure1_0("%get-soundcard-c-config-variable",
			    get_soundcard_c_config_variable);
	
	gh_new_procedure1_0("%get-mididev-c-config-variable",
			    get_mididev_c_config_variable);
	
	gh_new_procedure1_0("%get-mididevlist-c-config-variable",
			    get_mididevlist_c_config_variable);
	
	gh_new_procedure1_0("%get-string-c-config-variable",
			    get_string_c_config_variable);
	
	gh_new_procedure1_0("%get-charptr-c-config-variable",
			    get_charptr_c_config_variable);
	
	gh_new_procedure1_0("%get-charptrlist-c-config-variable",
			    get_charptrlist_c_config_variable);
	
	guileconfig::Config = this;
}

guileconfig::~guileconfig ()

{
	delete _app_name;
}
 
//////////////////////////////////////////////////////////////////////
//  guileconfig::bootstrap()
//  call this after constructing the guileconfig object to 
//  load user and system guile startup code. 
//////////////////////////////////////////////////////////////////////

int
guileconfig::bootstrap() 
{

	char path[PATH_MAX+1];
	char app_inst_root[PATH_MAX+1];
	char app_envt_var[PATH_MAX+1];
	char app_conffile_name[PATH_MAX+1];
	char app_conffile_envt_var[PATH_MAX+1];
	char *p;
	char *env;
	bool config_loaded = false;
	int i;

	p = _app_name;
	i = 0;
	while (*p) {
		app_envt_var[i++] = toupper(*p++);
	}
	app_envt_var[i] = 0;
	
	sprintf (app_conffile_name, ".%src", _app_name);
	
	p = app_conffile_name+1;
	i = 0;
	while (*p) {
		app_conffile_envt_var[i++] = toupper(*p++);    
	}
	app_conffile_envt_var[i] = 0;
	
	// find and load config-defs.scm and config-variables.scm
	// they get installed in $PREFIX/lib/$APPNAME/scm.
	// If the user has defined the $APPNAME env variable, use
	// that as the prefix instead. 
	
	if((env = getenv(app_envt_var)) != 0) {
		sprintf(app_inst_root, "%s/lib/guileconfig/scm", env);
	} else {
		strcpy(app_inst_root, GUILECONFIG_GUILEDIR);
	}
	
	// export this to guile as %guileconfig-scmdir 
	gh_define("%guileconfig-scmdir", gh_str02scm(app_inst_root));
	
	sprintf(path, "%s/bootstrap.scm", app_inst_root);
	if(access (path, R_OK) == 0) {
		gh_load(path);
	} else {
		error << "guileconfig : can't load Guile bootstrap "
			 "code from " 
		      << path
		      << endmsg;
		return -1;
	}
	
	// load the application's variable definition file

	sprintf(path, "%s/%s-variables.scm", app_inst_root, _app_name);
	if(access (path, R_OK) == 0) {
		gh_load(path);
	} else {
		error << "guileconfig : can't load configuration "
			 "variable definitions from " 
		      << path
		      << endmsg;
		return -1;
	}
	
	// now load a config file.  Environment variable spec has first
	// priority, then $HOME/.<APPNAME>rc, then ./.<APPNAME>rc.

	if ((env = getenv (app_conffile_envt_var)) != 0) {
		info << "guileconfig : found " << _app_name 
		     << " rc file in:\n\t " 
		     << env << endmsg;
		gh_load (env);
		config_loaded = true;

	} else if ((env = getenv ("HOME")) != 0) {
		sprintf (path, "%s/%s", env, app_conffile_name);
		
		if (access (path, R_OK) == 0) {
			info << "guileconfig : found " << _app_name 
			     << " rc file in:\n\t " 
			     << path << endmsg;
			gh_load (path);
			config_loaded = true;
		}

	} else if (access (app_conffile_name, R_OK) == 0) {
		info << "guileconfig : found " << _app_name 
		     << " rc file in:\n\t "
		     << app_conffile_name 
		     << endmsg;
		gh_load (app_conffile_name);
		config_loaded = true;
	}
	
	// FIXME : add command line processing here! -- bg
	
	if (!config_loaded) {
		error << "guileconfig : no configuration file found." 
		      << endmsg;
		return -1;
	} 
	
	return post_parse ();
}




